//
// Copyright (c) 2004-2011 StreamBase Systems, Inc. All rights reserved.
//
package com.streambase.contrib.adapter.sharedexample;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.adapter.InputAdapter;
import com.streambase.sb.adapter.OutputAdapter;
import com.streambase.sb.operator.Operator;
import com.streambase.sb.operator.Operator.SharedObject;
import com.streambase.sb.util.Msg;

public class MySharedObject implements Runnable, SharedObject
{
    
    private InputAdapter _inAdapter = null;
    private OutputAdapter _outAdapter;
    private Thread myThread=null;
    private volatile boolean shouldRun=true;
    private Object lock = new Object();
    
    //
    // Tracks the 'suspended' state of each adapter. We're not calling Operator.isSuspended() et. al. because
    // those are blocking calls and that can get us into trouble when we're in 3rd-party API callbacks...
    //
    private volatile boolean _inputAdapterIsSuspended = false;
    private volatile boolean _outputAdapterIsSuspended = false;
    private volatile boolean _outputAdapterIsShutdown = false;

    //
    // This value is obtained from the owning adapter, as specified in its Adapter Properties tab
    //
    private String settingFromAdapter;
    
    // Queue used to hold tuples the output Adapter wishes to pass on to the input adapter
    private LinkedBlockingQueue<Tuple> sendQ= new LinkedBlockingQueue<Tuple>();
    
    public MySharedObject(OutputAdapter outAdatper) { 
    	_outAdapter=outAdatper;
    }

    /**
     * Establishes the connection to the configured target(s)
     */
    public void run()
    {
        //
        // Use this if your shared object needs a thread for it's work
        //
    	Msg.debug(Msg.format("The shared object thread is starting"));
    	while (shouldRun) {
    		try {
				Thread.sleep(100);
			} catch (InterruptedException e) {}
    	}
    	Msg.debug(Msg.format("The shared object thread is exiting"));
    }

    /**
     * The following methods are defined by the SharedObject interface
     */
    
    /**
     * Start the SharedObject.  SharedObjects are started BEFORE any
     * Operators are started.
     * @throws StreamBaseException on error
     */
    public void startObject() throws StreamBaseException {
    	myThread=new Thread(this, _outAdapter.getName());
    	myThread.start();
    }
    
    /**
     * Resume the SharedObject. SharedObjects are resumed BEFORE any
     * Operators are resumed.
     * @throws StreamBaseException on error
     */
    public void resumeObject() throws StreamBaseException {
    	
    }
    
    /**
     * Suspend the SharedObject. SharedObjects are suspended AFTER any
     * Operators are suspended.
     * @throws StreamBaseException on error
     */
    public void suspendObject() throws StreamBaseException {
    	
    }
    
    /**
     * Shutdown the SharedObject. SharedObjects are shutdown AFTER any
     * Operators are shutdown.
     * @throws StreamBaseException on error
     */
    public void shutdownObject() throws StreamBaseException {
    	// The adapters are both shutdown by this point
    	
    	shutdownThread();
    }

    public void shutdownThread() {
    	shouldRun=false;
    	if (myThread != null) {
    		try {
				myThread.join(1000);
			} catch (InterruptedException e) {
				Msg.debug(Msg.format("The shared object thread did not join in a timely manner"));
			}
    	}
    }
    
    /**
     * Indicates that the given operator is resuming.
     * 
     * @param adapter the resuming operator.
     * @throws StreamBaseException 
     */
    public void resume(Operator adapter) throws StreamBaseException
    {
    	// Have the shared object track the state of both input and output adapters
        if (adapter instanceof InputAdapter)
        {
            _inputAdapterIsSuspended = false;
        }
        else
        {
            _outputAdapterIsSuspended = false;
        }
        
        //
        // This string will appear twice if the object is shared because
        // both In and Out adapters will call this method when their own
        // resume() callback is invoked...
        //
        System.err.println(String.format("SharedObject instance %s is used by the following adapter pair: In: %s, Out: %s, and the shared setting is '%s'.",
                                         this.toString(),
                                         _inAdapter == null ? "<null>" : _inAdapter.getFullyQualifiedName(),
                                         _outAdapter == null ? "<null>" : _outAdapter.getFullyQualifiedName(),
                                         getSettingFromAdapter()));
    }

    /**
     * Indicates that the given operator is suspending.
     * 
     * @param adapter the suspending operator.
     */
    public void suspend(Operator adapter)
    {
        if (adapter instanceof InputAdapter)
        {
            _inputAdapterIsSuspended = true;
        }
        else
        {
            _outputAdapterIsSuspended = true;
        }
    }
    
    public void shutdown(Operator adapter)
    {
        if (adapter instanceof OutputAdapter)
        {
        	_outputAdapterIsShutdown = true;
            _outAdapter = null;
            // In this example, when the output adapter is shutdown, the shared object is going to be shutdown and removed.
            shutdownThread();
        }
    }
    
    // The output adapter can put tuples on this queue that the input adapter can up pick
    public void sendTuple(Tuple t) throws StreamBaseException {
    	try {
			sendQ.put(t);
		} catch (InterruptedException e) {
			throw new StreamBaseException(e);
		}
    }
    
    // Wait for timeoutMS for a tuple.
    public Tuple waitForTuple(int timeoutMS) {
    	try {
			return sendQ.poll(timeoutMS, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			return null;
		}
    }
    
    public String getSettingFromAdapter()
    {
        return settingFromAdapter;
    }
    
    public void setSettingFromAdapter(String setting)
    {
        settingFromAdapter = setting;
    }
    
    /**
     * @return true if the input adapter used by this connection is currently suspended.
     */
    public boolean isInputAdapterSuspended()
    {
        return _inputAdapterIsSuspended;
    }
    
    /**
     * @return true if the output adapter used by this connection is currently suspended.
     */
    public boolean isOutputAdapterSuspended()
    {
        return _outputAdapterIsSuspended;
    }
 
    public boolean isOutputAdapterShutdown()
    {
        return _outputAdapterIsShutdown;
    }
    
    public InputAdapter getInputAdapter()
    {
        return _inAdapter;
    }
    
    public void setInputAdapter(InputAdapter adapter) throws StreamBaseException
    {
    	// This example only allows one input adapter to be linked to an output
    	// Since this method is called from the input resume() method, it must be synchronized
    	synchronized (lock) {
    		if (_inAdapter != null) {
    			throw new StreamBaseException(Msg.format("Shared object already has input adatper {0}, can not set with {1}",
    					_inAdapter.getName(), adapter.getName()));
    		}
    		this._inAdapter = adapter;
    	}
    }
    
    public OutputAdapter getOutputAdapter()
    {
        return _outAdapter;
    }
    
    public void setOutputAdapter(OutputAdapter adapter)
    {
        this._outAdapter = adapter;
    }
  
}
