//
// Copyright (c) 2004-2011 StreamBase Systems, Inc. All rights reserved.
//
package com.streambase.contrib.adapter.sharedexample;

import org.slf4j.Logger;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.adapter.InputAdapter;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.PortCounts;
import com.streambase.sb.operator.TypecheckException;

/**
 * For in-depth information on implementing a custom adapter, please
 * see "Developing StreamBase Embedded Adapters" in the StreamBase documentation.
 * <p>
 */
public class SharedInputAdapter extends InputAdapter implements Parameterizable, Runnable
{
    public static final long serialVersionUID = 1219377728750L;

    private Logger logger;
    // Properties
    private boolean isSharedAdapter;
    private String sharedAdapter;
    private String settingForSharedObject;
    private String originalSettingForSharedObject=null;
    private boolean firstResume=true;
    
    private MySharedObject mySharedObject=null;
    private Schema outputSchema=null;
    
    private String shortDisplayName = "Shared Input";
    private String displayName = "Shared Input Adapter Example";

    /**
     * The constructor is called when the Adapter instance is created, but before the Adapter 
     * is connected to the StreamBase application. We recommended that you set the initial input 
     * port and output port count in the constructor by calling setPortHints(inPortCount, outPortCount). 
     * The default is no input ports or output ports. The constructor may also set default values for 
     * adapter parameters. These values will be displayed in StreamBase Studio when a new adapter is 
     * dragged to the canvas, and serve as the default values for omitted optional parameters.
     */
    public SharedInputAdapter()
    {
        super();
        logger = getLogger();
        setPortHints(0, 1);
        setShortDisplayName(shortDisplayName);
        setDisplayName(displayName);
    }

    @Override
    public PortCounts getPortCounts() throws TypecheckException
    {
        return new PortCounts(0, 1);
    }

    /**
     * Typecheck this adapter. The typecheck method is called after the adapter instance is
     * connected in the StreamBase application. The adapter should validate its parameters 
     * and throw TypecheckException if any problems are found. The message associated with 
     * the thrown TypecheckException will be displayed in StreamBase Studio during authoring,
     * or printed on the console by sbd. Input adapters should set the schema0 of each output
     * port by calling the setOutputSchema(portNum, schema0) method for each output port. 
     * If the adapter needs to change the number of input ports based on parameter values,
     * it should call requireInputPortCount(portCount) at this point. 
     */
    public void typecheck() throws TypecheckException
    {
    	outputSchema = new Schema(null, new Schema.Field [] { new Schema.Field("anInt", CompleteDataType.forInt()), new Schema.Field("setting", CompleteDataType.forString()), } );
        setOutputSchema(0, outputSchema);
        
        //
        // If the adapter is Shared, make sure the name of the Shared output adapter is set
        //
        if (getIsSharedAdapter())
        {
            if (getSharedAdapter() == null || getSharedAdapter().trim().length() == 0)
            	throw new PropertyTypecheckException(SharedInputAdapterBeanInfo.IS_SHARED_PROPERTY_NAME,
            			"You must set 'Shared With The Output Adapter Named:' to the name of a 'Shared Output' adapter.");
        }
        else
        {
            //
            // 
            //
            if ((getSettingForSharedObject() == null || getSettingForSharedObject().trim().length() == 0))
            {
            	throw new PropertyTypecheckException(SharedInputAdapterBeanInfo.SHARED_SETTING_PROPERTY_NAME,
            			"If 'Shared With An Output Adapter' is not set, you must set 'Setting To Pass Along To The Shared Object'.");
            }
            else
            {
                //
                // TODO: Perform typecheck for "Setting To Pass Along To The Shared Object" value
                //
            }
        }
    }

    /**
     * Initialize the adapter. If typecheck succeeds, the init method is called before
     * the StreamBase application is started. Note that your adapter is not required to
     * define the init method, unless you need to register a runnable or perform
     * initialization of a resource such as, for example, a JDBC pool.
     */
    public void init() throws StreamBaseException
    {
        super.init();
        logger.debug(String.format("Input adapter %s: init()", getFullyQualifiedName()));
        
        // Register the object so it will be run as a thread managed by StreamBase.
        registerRunnable(this, true);
    }
    
     /**
     * Main thread of the adapter. At this point, the application begins to run. StreamBase will start threads for any
     * managed runnables registered by earlier calls to registerRunnable. Input adapters can call sendOutput(port,
     * tuple) at any time to output a tuple to the specified port.
     */
    public void run()
    {
    	logger.debug(String.format("Input adapter %s: run()", getFullyQualifiedName()));
    	Tuple t=outputSchema.createTuple();
    	int count=0;
        while (shouldRun())
        {
        	count++;
        	try {
				t.setInt(0, count);
				t.setString(1, getSettingForSharedObject());
				sendOutput(0, t);
			} catch (Exception e) {
				logger.warn(getName(), e);
			}
        	if ((mySharedObject != null) && mySharedObject.isOutputAdapterShutdown()) {
        		// If linked with the output adapter and it's shutdown, probably should do something different
        		// For this sample, just revert to being an unlinked adapter.
        		
        		// Restore the original setting for this input adapter
        		setSettingForSharedObject(originalSettingForSharedObject);
        		
        		mySharedObject=null;
        	}
            try
            {
            	int waitMS=5000;
            	if (mySharedObject != null) {
            		while (waitMS > 0) {
            			long start=System.nanoTime();
	            		Tuple tuple=mySharedObject.waitForTuple(waitMS);
	            		
	            		if (tuple != null) {
	            			try {
								sendOutput(0, tuple);
							} catch (StreamBaseException e) {
								logger.warn(getName(), e);
							}
	            		}
	            		long duration = (System.nanoTime() - start)/1000000;
	            		waitMS = waitMS - (int)duration; 
            		}
            	} else {
            		Thread.sleep(waitMS);
            	}
            }
            catch (InterruptedException e)
            {
            }
        }
    }

    /**
     * Shutdown adapter by cleaning up any resources used (e.g. close an output file
     * if it has been opened). When the application is shutting down, the adapter's shutdown.
     * method will be called first. Once this has returned, all threads should exit and the
     * adapter is considered shutdown.
     */
    public void shutdown()
    {
    	if (mySharedObject != null)
    		mySharedObject.shutdown(this);
    }
    /**
     * This method will be called when an adapter starts or resumes execution,
     * before any registered runnables are started or resumed.  Note that if the
     * application is shutdown directly from a suspended state, this will
     * <b>not</b> be called; instead {@link #shutdown()} will be called.
     *
     * resume() is a callback that is called by the StreamBase runtime.
     * @throws StreamBaseException 
     */
    public void resume() throws StreamBaseException
    {
    	logger.debug(String.format("Input adapter %s: resume()", getFullyQualifiedName()));
        if (firstResume && isSharingLinkToAdapter())
        {
        	firstResume=false;
        	// 
            // Now link with the output adapter.
            // Verify that we are indeed sharing the SharedObject with the right output adapter
            //
        	mySharedObject = (MySharedObject)getRuntimeEnvironment().getSharedObjectManager().getSharedObject(getSharedAdapter()) ;
        	
            if (mySharedObject == null) {
                String sMsg = String.format("Error trying to share connection of output adapter '%s' (reason: Could not find an output adapter named '%s' in the application.) -- stopping adapter.",
                                            getSharedAdapter(),
                                            getSharedAdapter());
                logger.warn(sMsg);
                throw new StreamBaseException(sMsg);
            }
            
            // If there is already an input adapter, this will throw an exception
            mySharedObject.setInputAdapter(this);
            //
            // Tell the shared object we've resumed
            //
            mySharedObject.resume(this);
            
            originalSettingForSharedObject=getSettingForSharedObject();
            // Now jam the setting from the output adapter
            setSettingForSharedObject(mySharedObject.getSettingFromAdapter());
        }
    }

    /**
     * This method will be called when an adapter suspends, before any
     * registered runnables are suspended.
     *
     * suspend() is a callback that is called by the StreamBase runtime.
     */
    public void suspend()
    {
        //
        // Tell the shared object we're suspended
        //
    	if (mySharedObject !=null)
    		mySharedObject.suspend(this);
    }

    /**
     * This method will be called by the StreamBase acceptor for each Tuple given to the adapter to process. Output
     * adapters should override this method to process tuples.
     *<p>
     * The default implementation does nothing.
     * 
     * @param inputPort
     *            the input port that the tuple is from (ports are zero based)
     * @param tuple
     *            the tuple from the given input port
     * @throws StreamBaseException
     *             Terminates the application.
     */
    public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException
    {

    }
   
    private boolean isSharingLinkToAdapter()
    {
        return getIsSharedAdapter() && getSharedAdapter() != null && getSharedAdapter().trim().length() > 0;
    }
    
    /***************************************************************************************
     * The getter and setter methods provided by the Parameterizable object.               *
     * StreamBase Studio uses them to determine the name and type of each property         *
     * and obviously, to set and get the property values.                                  *
     ***************************************************************************************/

    public void setIsSharedAdapter(boolean b)
    {
        this.isSharedAdapter = b;
    }

    public boolean getIsSharedAdapter()
    {
        return this.isSharedAdapter;
    }

    public void setSharedAdapter(String s)
    {
        this.sharedAdapter = s;
    }

    public String getSharedAdapter()
    {
        return this.sharedAdapter;
    }
    
    public boolean shouldEnableSharedAdapter()
    {
        return getIsSharedAdapter();
    }

    public void setSettingForSharedObject(String s)
    {
        this.settingForSharedObject = s;
    }

    public String getSettingForSharedObject()
    {
        return this.settingForSharedObject;
    }
    
    public boolean shouldEnableSettingForSharedObject()
    {
        return !getIsSharedAdapter();
    }
    
    public static void main(String [] args) throws TypecheckException
    {
        SharedInputAdapter lia = new SharedInputAdapter();
        lia.setSettingForSharedObject("Foo");
        lia.typecheck();
        System.out.println("Done!");
    }
}

