//
// Copyright (c) 2004-2011 StreamBase Systems, Inc. All rights reserved.
//
package com.streambase.contrib.adapter.sharedexample;

import org.slf4j.Logger;

import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.adapter.OutputAdapter;
import com.streambase.sb.operator.AlreadyRegisteredException;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.PortCounts;
import com.streambase.sb.operator.TypecheckException;
import com.streambase.sb.util.Msg;

/**
 * For in-depth information on implementing a custom adapter, please
 * see "Developing StreamBase Embedded Adapters" in the StreamBase documentation.
 * <p>
 */
public class SharedOutputAdapter extends OutputAdapter implements Parameterizable, Runnable
{
    public static final long serialVersionUID = 1219356627328L;

    private Logger logger;
    
    //
    // The shared object!
    //
    private MySharedObject mySharedObject;

    // Properties
    private String settingForSharedObject;

    private String shortDisplayName = "Shared Output";
    private String displayName = "Shared Output Adapter Example";
    
    /**
     * The constructor is called when the Adapter instance is created, but before the Adapter 
     * is connected to the StreamBase application. We recommended that you set the initial input 
     * port and output port count in the constructor by calling setPortHints(inPortCount, outPortCount). 
     * The default is no input ports or output ports. The constructor may also set default values for 
     * adapter parameters. These values will be displayed in StreamBase Studio when a new adapter is 
     * dragged to the canvas, and serve as the default values for omitted optional parameters.
     */
    public SharedOutputAdapter()
    {
        super();
        logger = getLogger();
        setPortHints(1, 0);
        setShortDisplayName(shortDisplayName);
        setDisplayName(displayName);
    }

    @Override
    public PortCounts getPortCounts() throws TypecheckException
    {
        return new PortCounts(1, 0);
    }

    /**
     * Typecheck this adapter. The typecheck method is called after the adapter instance is
     * connected in the StreamBase application. The adapter should validate its parameters 
     * and throw TypecheckException if any problems are found. The message associated with 
     * the thrown TypecheckException will be displayed in StreamBase Studio during authoring,
     * or printed on the console by sbd. Input adapters should set the schema of each output
     * port by calling the setOutputSchema(portNum, schema) method for each output port. 
     * If the adapter needs to change the number of input ports based on parameter values,
     * it should call requireInputPortCount(portCount) at this point. 
     */
    public void typecheck() throws TypecheckException
    {
        //
        // TODO: Perform proper typecheck for "Setting To Initialize Shared Object" value.
        // As an example here we'll just check for non-empty values.
        //
        if (getSettingForSharedObject() == null || getSettingForSharedObject().trim().length() == 0)
            throw new PropertyTypecheckException(SharedOutputAdapterBeanInfo.SHARED_SETTING_PROPERTY_NAME, "The 'Setting To Initialize Shared Object' must not be left blank.");

        Schema s=getInputSchema(0);
        if (!(s.getField(0).getDataType().equals(DataType.INT) && s.getField(1).getDataType().equals(DataType.STRING))) {
        	throw new TypecheckException(Msg.format("{0}: Input schema must start with an Int followed by a String", getName()));
        }
        requireInputPortCount(1);
    }
    
    /**
     * Initialize the adapter. If typecheck succeeds, the init method is called before
     * the StreamBase application is started. Note that your adapter is not required to
     * define the init method, unless you need to register a runnable or perform
     * initialization of a resource such as, for example, a JDBC pool.
     */
    public void init() throws StreamBaseException
    {
        super.init();
        logger.debug(String.format("Output adapter %s: init()", getFullyQualifiedName()));

        //
        // Perform necessary initialization of the shared object
        // This sequence relies on the fact that operator init() is known to be single threaded.
        // 
        SharedObjectManager sharedObjectManager=getRuntimeEnvironment().getSharedObjectManager();
        
        // We are going to use the unqualified operator name (canvas name) as the shared key
        String key=getName();
        mySharedObject = new MySharedObject(this);
        try {
            sharedObjectManager.registerSharedObject(key, mySharedObject);
        } catch (AlreadyRegisteredException e) {
        	// In this example, there can only be one output adapter and it owns the shared object
        	logger.debug(getName(), e);
        	throw new StreamBaseException(e);
        }
      
        mySharedObject.setSettingFromAdapter(getSettingForSharedObject());
        // Register this object so it will be run as a thread managed by StreamBase.
        registerRunnable(this, true);
    }

 
    /**
     * Main thread of the adapter. At this point, the application begins to run. StreamBase will start threads for any
     * managed runnables registered by earlier calls to registerRunnable. Input adapters can call sendOutput(port,
     * tuple) at any time to output a tuple to the specified port.
     */
    public void run()
    {
    	logger.debug(String.format("Output adapter %s: run()", getFullyQualifiedName()));

        while (shouldRun())
        {
            try
            {
                Thread.sleep(500);
            }
            catch (InterruptedException e)
            {
            }
        }
    }

    /**
     * Shutdown adapter by cleaning up any resources used (e.g. close an output file
     * if it has been opened). When the application is shutting down, the adapter's shutdown.
     * method will be called first. Once this has returned, all threads should exit and the
     * adapter is considered shutdown.
     */
    public void shutdown()
    {
        //
        // Allow the shared object to perform any shutdown operations it needs
        //
        mySharedObject.shutdown(this);
        SharedObjectManager sharedObjectManager=getRuntimeEnvironment().getSharedObjectManager();
        // This adapter is shutting down so remove it 
        sharedObjectManager.removeSharedObject(getName());
    }

    /**
     * This method will be called when an adapter starts or resumes execution,
     * before any registered runnables are started or resumed.  Note that if the
     * application is shutdown directly from a suspended state, this will
     * <b>not</b> be called; instead {@link #shutdown()} will be called.
     *
     * resume() is a callback that is called by the StreamBase runtime.
     * @throws StreamBaseException 
     */
    public void resume() throws StreamBaseException
    {
    	logger.info(String.format("Output adapter %s: resume()", getFullyQualifiedName()));
        //
        // Tell the shared object to resume 
        //
        mySharedObject.resume(this);
    }

    /**
     * This method will be called when an adapter suspends, before any
     * registered runnables are suspended.
     *
     * suspend() is a callback that is called by the StreamBase runtime.
     */
    public void suspend()
    {
        //
        // Tell the shared object to suspend (because we just got suspended) 
        //
        mySharedObject.suspend(this);
    }

    /**
     * This method will be called by the StreamBase server for each
     * Tuple given to the adapter to process.  Output adapters should
     * override this method to process tuples.
     *<p>
     * The default implementation does nothing.
     * 
     * @param inputPort the input port that the tuple is from (ports are zero based)
     * @param tuple the tuple from the given input port
     * @throws StreamBaseException Terminates the application. 
     */
    public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException
    {
    	if (mySharedObject.getInputAdapter() != null) {
    		// There's an input adapter sharing our object
    		try {
				mySharedObject.sendTuple(tuple.clone());
			} catch (CloneNotSupportedException e) {
				logger.warn(getName(), e);
			}
    	} else {
    		logger.info(Msg.format("{0}: No linked tuple, just dropping: {1}", getName(), tuple.toString()));
    	}
    	
    }

    /***************************************************************************************
     * The getter and setter methods provided by the Parameterizable object.               *
     * StreamBase Studio uses them to determine the name and type of each property         *
     * and obviously, to set and get the property values.                                  *
     ***************************************************************************************/

    public void setSettingForSharedObject(String s)
    {
        this.settingForSharedObject = s;
    }

    public String getSettingForSharedObject()
    {
        return this.settingForSharedObject;
    }
}
